%
% Author: Jan Lellmann, J.Lellmann@damtp.cam.ac.uk
% Date: 18/11/2013
%

    %% load image
    
    img = double(imread('grand-tetons.jpg'))/255.0;
    img = max(0,min(1,imresize(img, 1/16)));
    
    image(img);
    keyboard;
    
    lambda = 0.5;
    sz = [size(img,1) size(img,2)]; % dimensions of image
    m = size(img,3);                % # channels
  
    %% Mosek - academic license, CVX2

    % min max{x1,x2} s.t. x1 + 2*x2 >= 1, x(1) <= 1
    
    % standard form:
    %   minimize z s.t. z >= x1, z >= x2, x1 + 2*x2 >= 1, x(1) <= 1
    
    
    %   ...
    
    
%   Syntax:      [res] = msklpopt(c,a,blc,buc,blx,bux,param,cmd)
%  
%   Purpose:     Solves the optimization problem
%            
%                  min c'*x
%                  st. blc <= a*x <= buc
%                      blx <= x   <= bux
%  
%   Description: Required arguments.
%                  c      Is a vector.
%                  a      Is a (preferably sparse) matrix.   
%  
%                Optional arguments.
%                  blc    Lower bounds on constraints. 
%                  buc    Upper bounds on constraints. 
%                  blx    Lower bounds on variables.
%                  bux    Upper bounds on variables.
%                  param  New MOSEK parameters.
%                  cmd    MOSEK commands. 
%  
%                blc=[] and buc=[] means that the 
%                lower and upper bounds are plus and minus infinite
%                respectively. The same interpretation is used for
%                blx and bux. Note -inf is allowed in blc and blx.
%                Similarly, inf is allowed in buc and bux. 
%  
%   See also:  mosekopt, mskqpopt    

    res = msklpopt([0 0 1]',...
                   [1 2 0; -1 0 0; -1 0 1; 0 -1 1],...
                   [1 -1 0 0]',...
                   [+inf +inf +inf +inf]',...
                   [-inf -inf -inf]',...
                   [+inf +inf +inf]');
               
    res.sol.itr.xx(1:2)
    
    keyboard;
    
    
    
    
    
    %% CVX + SDPT3
    
    edit eliminate.m
    dbstop in eliminate.m at 24
    
    cvx_begin
        cvx_solver SDPT3
        variables x(2,1)
        minimize max(x(1),x(2))
        subject to
            x(1) + 2*x(2) >= 1
            x(1) <= 1
    cvx_end
    
    x
   
    keyboard
    
    %% CVX + Mosek
    
    cvx_begin
        cvx_solver Mosek % SeDuMi, Gurobi, ...
        variables x(2,1)
        minimize max(x(1),x(2))
        subject to
            x(1) + 2*x(2) >= 1
            x(1) <= 1
    cvx_end
    
    x
    
    keyboard

    %% Dual variables
    
    cvx_begin
        cvx_solver Mosek
        variables x(2,1)
        dual variable y1
        dual variable y2
        minimize max(x(1),x(2))
        subject to
            y1: x(1) + 2*x(2) >= 1
            y2: x(1) <=1
    cvx_end
    
    x
    y1 % active constraint! (complementarity condition)
    y2 % (not) active? (change x(1) >= 1)
    
    keyboard

    %% Global settings 
    
    cvx_solver Mosek
    cvx_precision low   
    
    % Solver goal:            precision <= 1.35e-006
    % "Solved" if             precision <= 1.22e-004
    % "Inaccurate/Solved" if  1.22e-004 < precision <= 1.22e-004    
    
    %%  ROF
    
    Grad = diffopn([size(img,1) size(img,2)],size(img,3),'forward','neumann');
    cvx_begin
        cvx_solver Mosek
        variables u(size(img))
        minimize sum((u(:) - img(:)).^2) + lambda * sum(abs(Grad * u(:)))
    cvx_end
    
    image(u)
    
    keyboard
    
    %% TV-L1
    
    lambda = 2.0;
    cvx_begin
        cvx_solver Mosek
        variables u(size(img))
        minimize sum(abs(u(:) - img(:))) + lambda * sum(abs(Grad * u(:)))
    cvx_end
    
    image(u)
    
    keyboard
    
    %% TV-L1 with variables
    
    cvx_begin
        cvx_solver Mosek
        variables u(size(img))
        
        gu = Grad * u(:);
        gu
        sum(abs(gu))
        
        minimize sum(abs(u(:) - img(:))) + lambda * sum(abs(gu))
    cvx_end
    
    keyboard
    
    %% TV-L1 strings vs. expressions
    
    cvx_begin
        cvx_solver Mosek
        variables u(size(img))
        
        gu = reshape(Grad * u(:), [prod(sz) 2*m]);
        
        absgu = reshape(sum(abs(gu),2), sz);
        
        dataterm = sum(abs(u(:) - img(:)));
        regularizer = sum(absgu(:));
       
        f = dataterm + lambda * regularizer;
        
        minimize(f)
    cvx_end
    
    dataterm
    regularizer
    
    image(u)
    
    keyboard;
    
    figure;
    imagesc(absgu);
    colorbar;
    
    keyboard;
    
    %% TV-L1 with better norm

    % does not work - disciplined convex programming error
    %cvx_begin
    %    variables u(size(img))
    %    
    %    gu = reshape(Grad * u(:), prod(sz), numel(sz)*m);
    %    
    %    regularizer = sum(sqrt(sum(gu.^2,2))); % L2 norm
    %    
    %    minimize sum(abs(u(:) - img(:))) + lambda * regularizer
    %cvx_end
    %
    %image(u)

    %% TV-L1 with better norm (correct)
    
    cvx_begin
        cvx_solver Mosek
        variables u(size(img))
        
        gu = reshape(Grad * u(:), prod(sz), numel(sz)*m);
        
        regularizer = sum(norms(gu,2,2)); % L2 norm - built-in function
        
        minimize sum(abs(u(:) - img(:))) + lambda * regularizer
    cvx_end
    
    image(u)
    
    keyboard   

    %% TV-L1 with functions
    
    cvx_begin
        cvx_solver Mosek
        variables u(size(img))
        
        minimize sum(abs(u(:) - img(:))) + lambda * TV(u)
    cvx_end
    
    image(u)
    
    keyboard
    
    %% TV-L1 inpainting with constraints
    
    [yy,xx,zz] = ndgrid(0:(1/(sz(1)-1)):1,0:(1/(sz(2)-1)):1,1:m);
    inpaint_region = abs(xx - 0.5) < 0.25 & abs(yy - 0.25) < 0.1;
    
    cvx_begin
        cvx_solver Mosek
        variables u(size(img))
        
        minimize TV(u)
        
        subject to
            u(~inpaint_region) == img(~inpaint_region)
    cvx_end
    
    image(u)
    
    keyboard
    
    %% higher-order

    lambda = 1.0;
    cvx_begin
        cvx_solver Mosek
        variables u(size(img))
        
        minimize sum(abs(u(:) - img(:))) + lambda * TV2(u)
    cvx_end
    
    image(u)
    
    keyboard

    %% segmentation
    c1 = [0 0 1]; c2 = [0 1 1];
    imr = reshape(img,prod(sz),m);
    f = norms(imr - ones(size(imr,1),1) * c1,2,2) - norms(imr - ones(size(imr,1),1) * c2,2,2);
    
    cvx_begin
        cvx_solver Mosek
        variables u(sz)
        
        minimize sum(u(:) .* f(:)) + 0.5 * TV(u)
        subject to
            0 <= u <= 1
    cvx_end
    
    imagesc(u)
    
    keyboard
    
    %% inf-convolution
    cvx_begin
        variables u(size(img))
        variables u1(size(u))
        variables u2(size(u))

        minimize sum(abs(u(:) - img(:))) + 2.0 * TV(u1) + 2.0 * TV2(u2)
        subject to
            u1 + u2 == u
    cvx_end
    
    imshow([img u; u1 u2]);

    keyboard
    
    %% inf-convolution extra fancy
    cvx_begin
        cvx_solver Mosek
        variables u(size(img))
        
        minimize sum(abs(u(:) - img(:))) + inf_conv(u, 2.0, 2.0)
    cvx_end
    
    image(u)
    
    keyboard

    %% modified inf-convolution
    Grad = diffopn(sz,m,'forward','neumann');
    Grad2 = diffopn(sz,numel(sz)*m,'backward','dirichlet');
    cvx_begin
        cvx_solver Mosek
        variables u(size(img))
        variables u1(size(u))
        variables u2(size(u))
        variables s1([size(u,1) size(u,2) 2 size(u,3)])
        
        v1 = reshape(Grad * u1(:) - s1(:), size(u,1)*size(u,2), 2*size(u,3));
        v2 = reshape(Grad2 * (Grad * u2(:) + s1(:)), size(u,1)*size(u,2), 2*2*size(u,3));

        minimize sum(abs(u(:) - img(:))) + 2.0 * sum(norms(v1,2,2)) + 2.0 * sum(norms(v2,2,2))
        subject to
            u1 + u2 == u
            Grad' * s1(:) == 0
    cvx_end
    
    imshow([img u; u1 u2]);
    
    %keyboard


    
    %%%%%
    
    

    %% optimization with integer constraints
    c1 = [0 0 1]; c2 = [0 1 1];
    imr = reshape(img,prod(sz),m);
    f = norms(imr - ones(size(imr,1),1) * c1,2,2) - norms(imr - ones(size(imr,1),1) * c2,2,2);
    
    cvx_begin
        cvx_solver Mosek
        variable u(sz) binary
        
        minimize sum(u(:) .* f(:)) + 0.5 * TV(u)
    cvx_end
    
    imagesc(u)   
    
    keyboard

    %% semidefinite programming
    
    % Euclidean distance matrices: D_ij = d(i,j) = ||x^i - x^j||_2.
    %
    % n x n squared Euclidean distances matrices X are exactly the matrices
    % of the form B(Y) with Y (n-1)x(n-1) positive semidefinite matrix.
    % B is a linear bijection.
    
    D = [0 1 3; 1 0 1; 3 1 0];
    n = size(D,1);
    V = null(ones(1,n));
    
    cvx_begin
        variable Y(n-1,n-1) symmetric
        minimize norm(reshape(D.^2 - B(V,Y),n*n,1),2)
        subject to
            Y == semidefinite(n-1)
    cvx_end

    X = B(V,Y);
    sqrt(X)

    keyboard

    %% semidefinite programming fancy
    
    D = [0 1 3; 1 0 1; 3 1 0];
    n = size(D,1);
    V = null(ones(1,n))
    
    cvx_begin
        variable X(n,n) symmetric
        minimize norm(reshape(D.^2 - X,n*n,1),2)
        subject to
            X == B(V,semidefinite(n-1))
    cvx_end
    
    sqrt(X)

    keyboard
    